#!/usr/bin/env python

################################################################################
# This script retrieves reads that either contain all-PAKs or all the bases are
# covered by PAKs and output the reads as FASTA
#
# Inputs:
#     [-m]: mode of retrieval (all-paks or all-bases)
#     [-k]: the list of PAKs
#     [-A]: mate A FASTA file
#     [-B]: mate B FASTA file
#     [-o]: output header
#     [-r]: check also the reverse complements of the PAKs (default=False)
#     [-v]: be verbose
# Outputs:
#     1. output_header_single.fasta: FASTA file containing only the reads that
#     pass the threshold criteria. A mate read is not included if it did not
#     meet the criteria.
#     2. output_head_A/B.fasta: FASTA files containing the read pairs if either
#     mate passes the threshold.
#
# Usage: find_PARs.py -k <K-mer_file> -A <mate_A.fasta> -B <mate_B.fasta>
#     -o <output_header> [-r] [-v] [-t <threshold>] [-c <covered>]
#
# Hong Ching Lee     06/07/2011
# Philipp Bayer 2014: removed creation of A and B fasta output
################################################################################

# import libraries
import argparse # parse input options
import sys
from Bio import Seq


def read_mers_to_dict(file_name, rev=False, is_verbose=False):
    '''This method reads all the K-mers from the input file into a dict.
    If rev=True, the reverse complement of each K-mer is also generated and
    taken into account in the downstream analysis.'''
    mers = {}
    mer_len = 0

    with open(file_name, "rU") as k_in:
        for line in k_in:
            columns = line.rstrip().split('\t')
            mer = (columns[0]).upper()

            if mer_len == 0:
                mer_len = len(mer)

            if is_verbose and mer in mers:
                print "Warning: K-mer", mer, "duplicated in input! Input K-mers potentially corrupted!"
            mers[mer] = int(columns[1])

            if rev:
                rev_mer = (Seq.reverse_complement(mer)).upper()
                if is_verbose and rev_mer in mers:
                    print "Warning: Reverse K-mer", mer, "duplicated in input! Input K-mers potentially corrupted!"
                mers[rev_mer] = int(columns[1])
    k_in.close()

    if is_verbose:
        if rev and (mer_len % 2) == 0:
            print "Warning: K-mer forward and reverse complement may be the same because K-mer length is even. E.g. AATT"
        print "K-mer file parsing DONE!!!"

    return (mers, mer_len)


def find_PARs(mers, mode, mer_len, A_file_name, B_file_name, out_single_name, out_A_name, out_B_name, is_verbose):
    '''This method retrieves reads from A and B_files if the reads meet the
    criterion (either all-paks or all-bases mode).'''

    in_A = open(A_file_name, "rU")
    in_B = open(B_file_name, "rU")
    out_single = open(out_single_name, "w")
    mer_bits = init_mer_bits(mer_len)
    all_paks_mode = False
    all_bases_mode = False

    if mode == "all-paks":
        all_paks_mode = True
    elif mode == "all-bases":
        all_bases_mode = True
    else:
        raise IOError("Unknown mode: " + mode)

    while (True):
        a_line1 = in_A.readline().rstrip()
        a_line2 = (in_A.readline().rstrip()).upper()
        b_line1 = in_B.readline().rstrip()
        b_line2 = (in_B.readline().rstrip()).upper()

        if not a_line2:
            break

        if (a_line1[0] != '>' or b_line1[0] != '>'):
            raise IOError("Expecting FASTA format inputs with one header line followed by one line containing the read sequence (reads spanning multiple lines are not supported)")

        print_A = False
        print_B = False

        if all_paks_mode:
            print_A = check_mers(a_line2, mers, mer_len, is_verbose)
            print_B = check_mers(b_line2, mers, mer_len, is_verbose)
        elif all_bases_mode:
            print_A = check_base_coverage(a_line2, mers, mer_len, mer_bits, is_verbose)
            print_B = check_base_coverage(b_line2, mers, mer_len, mer_bits, is_verbose)


        # Output all reads meeting the threshold to a single FASTA file
        if print_A:
            out_single.write(">" + a_line1[1:] + "\n")
            out_single.write(a_line2 + "\n")
        if print_B:
            out_single.write(">" + b_line1[1:] + "\n")
            out_single.write(b_line2 + "\n")

    in_A.close()
    in_B.close()
    out_single.close()


def check_mers(seq, mers, mer_len, is_verbose):
    '''This method goes through a read sequence and checks each constituent
    k-mer is in the mers dict (that is, it is a PAK). Returns True if all k-mers
    in the read are PAKs, else returns False.'''
    n_mers = len(seq) - mer_len + 1 # number of K-mers in this read

    for i in range(0, n_mers):
        if seq[i:i + mer_len] not in mers:
            return False

    return True


def check_base_coverage(seq, mers, mer_len, mer_bits, is_verbose):
    '''This method goes through a read sequence and checks if each base is
    covered by some k-mer(s) specified in the mers dict. Returns True if all
    bases are covered, else returns False.'''
    n_mers = len(seq) - mer_len + 1 # number of K-mers in this read
    covered = 0 # bit-wise recording if bases are covered by mers

    for i in range(0, n_mers):
        if seq[i:i + mer_len] in mers:
            cur_mer_bits = int(mer_bits) << (n_mers - i - 1)
            covered = covered | cur_mer_bits

    if is_verbose:
        print "Bit-wise base coverage:", bin(covered)

    if count_sparse_ones(covered) != len(seq): # count number of 1's and compare with read length to see if all bases are covered
        return False
    else:
        return True


def init_mer_bits(mer_len):
    '''This method creates a binary number with mer_len number of 1's'''
    mer_bits = 0

    for i in range(0, mer_len):
        mer_bits = (mer_bits << 1) + 1

    return mer_bits


def count_sparse_ones(num):
    '''This method counts the number of 1's in num'''
    total = 0

    while num:
        total += 1
        num = num & (num - 1)

    return total


def check_param(parser, args):
    '''This method checks the input parameters.'''

    if (args.kmer_file_name is None):
        print("===Please specify input PAK file.===")
        parser.print_help()
    elif (args.A_file_name is None):
        print("===Please specify input FASTQ reads A file.===")
        parser.print_help()
    elif (args.B_file_name is None):
        print("===Please specify input FASTQ reads B file.===")
        parser.print_help()
    elif (args.output_name is None):
        print("===Please specify output name. E.g. -o BnaA_03_001===")
        parser.print_help()
    else:
        return True

    return False


if __name__ == '__main__':

    # set input options
    usage = "%(prog)s -k <K-mer_file> -A <mate_A.fasta> -B <mate_B.fasta> -o <output_header> [-r] [-v] [-m <mode>]\n\n"
    descriptions = "For PAV Pipeline: retrieve PAV reads from FASTA files.\nNote: K-mer_file should be" +\
        " in jellyfish count output format (K-mer\tcount) per line."
    parser = argparse.ArgumentParser(usage=usage, description=descriptions)
    parser.add_argument("-m", "--mode", dest="mode", help="Method for identifying PARs: all-paks or all-bases [default: %(default)s]", choices=["all-paks", "all-bases"], default="all-paks")
    parser.add_argument("-k", "--kmers", dest="kmer_file_name", required=True, help="input the file containing the PAKs")
    parser.add_argument("-A", dest="A_file_name", required=True, help="input the FASTA mate A file")
    parser.add_argument("-B", dest="B_file_name", required=True, help="input the FASTA mate B file")
    parser.add_argument("-o", dest="output_name", required=True, help="output file header.")
    parser.add_argument("-r", "--reverse", dest="rev", action="store_false", help="search reverse complement [default=%(default)s]", default=True)
    parser.add_argument("-v", "--verbose", dest="is_verbose", action="store_true", help="be verbose [default=%(default)s]", default=False)

    args = parser.parse_args()

    if check_param(parser, args):
        (mers, mer_len) = read_mers_to_dict(args.kmer_file_name, args.rev, args.is_verbose)
        out_single_name = args.output_name + "_single.fasta"
        out_A_name      = args.output_name + "_A.fasta"
        out_B_name      = args.output_name + "_B.fasta"

        if args.is_verbose:
            print "Input K-mer PAK file name: " + str(args.kmer_file_name)
            print "K-mer length: " + str(mer_len)
            print "Fasta file name for all single matches: " + out_single_name
            print "Fasta file name for outputting reads A when EITHER read A or B matches the number of PAKs threshold requirement: " + out_A_name
            print "Fasta file name for outputting reads B when EITHER read A or B matches the number of PAKs threshold requirement: " + out_B_name

        find_PARs(mers, args.mode, mer_len, args.A_file_name, args.B_file_name, out_single_name, out_A_name, out_B_name, args.is_verbose)
